@@ -8,6 +8,8 @@
  - Licensed under the GNU GPL version 3 or higher.
  -}
 
+{-# LANGUAGE CPP #-}
+
 module Git (
 	Repo(..),
 	Ref(..),
@@ -30,11 +32,15 @@ module Git (
 ) where
 
 import Network.URI (uriPath, uriScheme, unEscapeString)
+#ifndef mingw32_HOST_OS
 import System.Posix.Files
+#endif
 
 import Common
 import Git.Types
+#ifndef mingw32_HOST_OS
 import Utility.FileMode
+#endif
 
 {- User-visible description of a git repo. -}
 repoDescribe :: Repo -> String
@@ -127,4 +133,8 @@ hookPath script repo = do
 	ifM (catchBoolIO $ isexecutable hook)
 		( return $ Just hook , return Nothing )
   where
+#if mingw32_HOST_OS
+	isexecutable f = doesFileExist f
+#else
 	isexecutable f = isExecutable . fileMode <$> getFileStatus f
+#endif
