@@ -25,6 +25,7 @@ import qualified Github.Repos as Github
 import qualified Github.Repos.Forks as Github
 import qualified Github.PullRequests as Github
 import qualified Github.Repos.Watching as Github
+import qualified Github.Repos.Starring as Github
 import qualified Github.Data.Definitions as Github ()
 import qualified Github.Issues as Github
 import qualified Github.Issues.Comments
@@ -42,16 +43,24 @@ import qualified Git.Branch
 import qualified Git.UpdateIndex
 import Git.HashObject
 import Git.FilePath
+import Git.CatFile
 import Utility.Env
 
 -- A github user and repo.
 data GithubUserRepo = GithubUserRepo String String
 	deriving (Eq, Show, Read, Ord)
 
-toGithubUserRepo :: Github.Repo -> GithubUserRepo
-toGithubUserRepo r = GithubUserRepo 
-	(Github.githubOwnerLogin $ Github.repoOwner r)
-	(Github.repoName r)
+class ToGithubUserRepo a where
+	toGithubUserRepo :: a -> GithubUserRepo
+
+instance ToGithubUserRepo Github.Repo where
+	toGithubUserRepo r = GithubUserRepo 
+		(Github.githubOwnerLogin $ Github.repoOwner r)
+		(Github.repoName r)
+
+instance ToGithubUserRepo Github.RepoRef where
+	toGithubUserRepo (Github.RepoRef owner name) = 
+		GithubUserRepo (Github.githubOwnerLogin owner) name
 
 repoUrl :: GithubUserRepo -> String
 repoUrl (GithubUserRepo user remote) =
@@ -80,8 +89,11 @@ requestName (RequestNum name _ _) = name
 data BackupState = BackupState
 	{ failedRequests :: S.Set Request
 	, retriedRequests :: S.Set Request
+	, retriedFailed :: S.Set Request
 	, gitRepo :: Git.Repo
 	, gitHubAuth :: Maybe Github.GithubAuth
+	, deferredBackups :: [Backup ()]
+	, catFileHandle :: Maybe CatFileHandle
 	}
 
 {- Our monad. -}
@@ -97,7 +109,7 @@ newtype Backup a = Backup { runBackup :: StateT BackupState IO a }
 inRepo :: (Git.Repo -> IO a) -> Backup a
 inRepo a = liftIO . a =<< getState gitRepo
 
-failedRequest :: Request -> Github.Error-> Backup ()
+failedRequest :: Request -> Github.Error -> Backup ()
 failedRequest req e = unless ignorable $ do
 	set <- getState failedRequests
 	changeState $ \s -> s { failedRequests = S.insert req set }
@@ -115,15 +127,16 @@ runRequest req = do
 type Storer = Request -> Backup ()
 data ApiListItem = ApiListItem ApiName Storer Bool
 apiList :: [ApiListItem]
-apiList = 
-	[ ApiListItem "userrepo" userrepoStore True
-	, ApiListItem "watchers" watchersStore True
+apiList =
+	[ ApiListItem "watchers" watchersStore True
+	, ApiListItem "stargazers" stargazersStore True
 	, ApiListItem "pullrequests" pullrequestsStore True
 	, ApiListItem "pullrequest" pullrequestStore False
 	, ApiListItem "milestones" milestonesStore True
 	, ApiListItem "issues" issuesStore True
 	, ApiListItem "issuecomments" issuecommentsStore False
-	-- comes last because it recurses on to the forks
+	-- Recursive things last.
+	, ApiListItem "userrepo" userrepoStore True
 	, ApiListItem "forks" forksStore True
 	]
 
@@ -131,7 +144,7 @@ apiList =
 api :: M.Map ApiName Storer
 api = M.fromList $ map (\(ApiListItem n s _) -> (n, s)) apiList
 
-{- List of toplevel api calls that are followed to get all data. -}
+{- List of toplevel api calls that are followed to get data. -}
 toplevelApi :: [ApiName]
 toplevelApi = map (\(ApiListItem n _ _) -> n) $
 	filter (\(ApiListItem _ _ toplevel) -> toplevel) apiList
@@ -142,35 +155,33 @@ lookupApi req = fromMaybe bad $ M.lookup name api
 	name = requestName req
 	bad = error $ "internal error: bad api call: " ++ name
 
-userrepoStore :: Storer
-userrepoStore = simpleHelper (noAuth Github.userRepo) $ \req r -> do
-	when (Github.repoHasWiki r == Just True) $
-		updateWiki $ toGithubUserRepo r
-	store "repo" req r
-
 watchersStore :: Storer
-watchersStore = simpleHelper (noAuth Github.watchersFor) $
+watchersStore = simpleHelper "watchers" Github.watchersFor' $
 	storeSorted "watchers"
 
+stargazersStore :: Storer
+stargazersStore = simpleHelper "stargazers" Github.stargazersFor $
+	storeSorted "stargazers"
+
 pullrequestsStore :: Storer
-pullrequestsStore = simpleHelper Github.pullRequestsFor' $
+pullrequestsStore = simpleHelper "pullrequest" Github.pullRequestsFor' $
 	forValues $ \req r -> do
 		let repo = requestRepo req
 		let n = Github.pullRequestNumber r
 		runRequest $ RequestNum "pullrequest" repo n
 
 pullrequestStore :: Storer
-pullrequestStore = numHelper Github.pullRequest' $ \n ->
+pullrequestStore = numHelper "pullrequest" Github.pullRequest' $ \n ->
 	store ("pullrequest" </> show n)
 
 milestonesStore :: Storer
-milestonesStore = simpleHelper Github.Issues.Milestones.milestones' $
+milestonesStore = simpleHelper "milestone" Github.Issues.Milestones.milestones' $
 	forValues $ \req m -> do
 		let n = Github.milestoneNumber m
 		store ("milestone" </> show n) req m
 
 issuesStore :: Storer
-issuesStore = withHelper (\a u r y ->
+issuesStore = withHelper "issue" (\a u r y ->
 	Github.issuesForRepo' a u r (y <> [Github.Open])
 		>>= either (return . Left)
 			(\xs -> Github.issuesForRepo' a u r
@@ -186,18 +197,23 @@ issuesStore = withHelper (\a u r y ->
 		runRequest (RequestNum "issuecomments" repo n)
 
 issuecommentsStore :: Storer
-issuecommentsStore = numHelper Github.Issues.Comments.comments' $ \n ->
+issuecommentsStore = numHelper "issuecomments" Github.Issues.Comments.comments' $ \n ->
 	forValues $ \req c -> do
 		let i = Github.issueCommentId c
 		store ("issue" </> show n ++ "_comment" </> show i) req c
 
+userrepoStore :: Storer
+userrepoStore = simpleHelper "repo" Github.userRepo' $ \req r -> do
+	store "repo" req r
+	when (Github.repoHasWiki r == Just True) $
+		updateWiki $ toGithubUserRepo r
+	maybe noop addFork $ Github.repoParent r
+	maybe noop addFork $ Github.repoSource r
+
 forksStore :: Storer
-forksStore = simpleHelper (noAuth Github.forksFor) $ \req fs -> do
+forksStore = simpleHelper "forks" Github.forksFor' $ \req fs -> do
 	storeSorted "forks" req fs
-	mapM_ (traverse . toGithubUserRepo) fs
-  where
-	traverse fork = whenM (addFork fork) $
-		gatherMetaData fork
+	mapM_ addFork fs
 
 forValues :: (Request -> v -> Backup ()) -> Request -> [v] -> Backup ()
 forValues handle req vs = forM_ vs (handle req)
@@ -208,39 +224,66 @@ type ApiNum v = ApiWith v Int
 type Handler v = Request -> v -> Backup ()
 type Helper = Request -> Backup ()
 
-noAuth :: (String -> String -> IO a) -> Maybe Github.GithubAuth -> String -> String -> IO a
-noAuth a _auth user repo = a user repo
+simpleHelper :: FilePath -> ApiCall v -> Handler v -> Helper
+simpleHelper dest call handle req@(RequestSimple _ (GithubUserRepo user repo)) =
+	deferOn dest req $ do
+		auth <- getState gitHubAuth
+		either (failedRequest req) (handle req) =<< liftIO (call auth user repo)
+simpleHelper _ _ _ r = badRequest r
 
-simpleHelper :: ApiCall v -> Handler v -> Helper
-simpleHelper call handle req@(RequestSimple _ (GithubUserRepo user repo)) = do
-	auth <- getState gitHubAuth
-	either (failedRequest req) (handle req) =<< liftIO (call auth user repo)
-simpleHelper _ _ r = badRequest r
+withHelper :: FilePath -> ApiWith v b -> b -> Handler v -> Helper
+withHelper dest call b handle req@(RequestSimple _ (GithubUserRepo user repo)) =
+	deferOn dest req $ do
+		auth <- getState gitHubAuth
+		either (failedRequest req) (handle req) =<< liftIO (call auth user repo b)
+withHelper _ _ _ _ r = badRequest r
 
-withHelper :: ApiWith v b -> b -> Handler v -> Helper
-withHelper call b handle req@(RequestSimple _ (GithubUserRepo user repo)) = do
-	auth <- getState gitHubAuth
-	either (failedRequest req) (handle req) =<< liftIO (call auth user repo b)
-withHelper _ _ _ r = badRequest r
-
-numHelper :: ApiNum v -> (Int -> Handler v) -> Helper
-numHelper call handle req@(RequestNum _ (GithubUserRepo user repo) num) = do
-	auth <- getState gitHubAuth
-	either (failedRequest req) (handle num req) =<< liftIO (call auth user repo num)
-numHelper _ _ r = badRequest r
+numHelper :: FilePath -> ApiNum v -> (Int -> Handler v) -> Helper
+numHelper dest call handle req@(RequestNum _ (GithubUserRepo user repo) num) =
+	deferOn dest req $ do
+		auth <- getState gitHubAuth
+		either (failedRequest req) (handle num req) =<< liftIO (call auth user repo num)
+numHelper _ _ _ r = badRequest r
 
 badRequest :: Request -> a
 badRequest r = error $ "internal error: bad request type " ++ show r
 
+{- When the specified file or directory already exists in git, the action
+ - is deferred until later. -}
+deferOn :: FilePath -> Request -> Backup () -> Backup ()
+deferOn f req a = ifM (ingit $ storeLocation f req)
+	( changeState $ \s -> s { deferredBackups = a : deferredBackups s }
+	, a
+	)
+  where
+	ingit f' = do
+		h <- getCatFileHandle
+		liftIO $ isJust <$> catObjectDetails h
+			(Git.Types.Ref $ show branchname ++ ":" ++ f')
+
+getCatFileHandle :: Backup CatFileHandle
+getCatFileHandle = go =<< getState catFileHandle
+  where
+  	go (Just h) = return h
+	go Nothing = do
+		h <- withIndex $ inRepo catFileStart
+		changeState $ \s -> s { catFileHandle = Just h }
+		return h
+
 store :: Show a => FilePath -> Request -> a -> Backup ()
 store filebase req val = do
-	file <- location (requestRepo req) <$> workDir
+	file <- (</>)
+		<$> workDir
+		<*> pure (storeLocation filebase req)
 	liftIO $ do
 		createDirectoryIfMissing True (parentDir file)
 		writeFile file (ppShow val)
+
+storeLocation :: FilePath -> Request -> FilePath
+storeLocation filebase = location . requestRepo
   where
-	location (GithubUserRepo user repo) workdir =
-		workdir </> user ++ "_" ++ repo </> filebase
+	location (GithubUserRepo user repo) =
+		user ++ "_" ++ repo </> filebase
 
 workDir :: Backup FilePath
 workDir = (</>)
@@ -307,13 +350,13 @@ commitWorkDir = do
 				-- Stage workDir files into the index.
 				h <- hashObjectStart r
 				Git.UpdateIndex.streamUpdateIndex r
-					[genstream r dir h]
+					[genstream dir h]
 				hashObjectStop h
 				-- Commit
 				void $ Git.Branch.commit "github-backup" fullname [branchref] r
 				removeDirectoryRecursive dir
   where
-  	genstream r dir h streamer = do
+  	genstream dir h streamer = do
 		fs <- filter (not . dirCruft) <$> dirContentsRecursive dir
 		forM_ fs $ \f -> do
 			sha <- hashFile h f
@@ -375,16 +418,16 @@ updateWiki fork =
 	remoteFor (GithubUserRepo user repo) =
 		"github_" ++ user ++ "_" ++ repo ++ ".wiki"
 
-addFork :: GithubUserRepo -> Backup Bool
-addFork fork =
-	ifM (elem fork <$> gitHubRemotes)
-		( return False
-		, do
-			liftIO $ putStrLn $ "New fork: " ++ repoUrl fork
-			_ <- addRemote (remoteFor fork) (repoUrl fork)
-			return True
-		)
+addFork :: ToGithubUserRepo a => a -> Backup ()
+addFork forksource = unlessM (elem fork <$> gitHubRemotes) $ do
+	liftIO $ putStrLn $ "New fork: " ++ repoUrl fork
+	void $ addRemote (remoteFor fork) (repoUrl fork)
+	gitRepo' <- inRepo $ Git.Config.reRead
+	changeState $ \s -> s { gitRepo = gitRepo' }
+
+	gatherMetaData fork
   where
+  	fork = toGithubUserRepo forksource
 	remoteFor (GithubUserRepo user repo) = "github_" ++ user ++ "_" ++ repo
 
 {- Adds a remote, also fetching from it. -}
@@ -412,8 +455,6 @@ fetchRepo :: Git.Repo -> Backup Bool
 fetchRepo repo = inRepo $ Git.Command.runBool
 	[Param "fetch", Param $ fromJust $ Git.Types.remoteName repo]
 
-{- Gathers metadata for the repo. Retuns a list of files written
- - and a list that may contain requests that need to be retried later. -}
 gatherMetaData :: GithubUserRepo -> Backup ()
 gatherMetaData repo = do
 	liftIO $ putStrLn $ "Gathering metadata for " ++ repoUrl repo ++ " ..."
@@ -433,7 +474,7 @@ loadRetry r = maybe [] (fromMaybe [] . readish)
 retryFile :: Git.Repo -> FilePath
 retryFile r = Git.localGitDir r </> "github-backup.todo"
 
-retry :: Backup (S.Set Request)
+retry :: Backup ()
 retry = do
 	todo <- inRepo loadRetry
 	unless (null todo) $ do
@@ -441,12 +482,11 @@ retry = do
 			"Retrying " ++ show (length todo) ++
 			" requests that failed last time..."
 		mapM_ runRequest todo
-	retriedfailed <- getState failedRequests
 	changeState $ \s -> s
-		{ failedRequests = S.empty
+		{ retriedFailed = failedRequests s
+		, failedRequests = S.empty
 		, retriedRequests = S.fromList todo
 		}
-	return retriedfailed
 
 summarizeRequests :: [Request] -> [String]
 summarizeRequests = go M.empty
@@ -460,77 +500,119 @@ summarizeRequests = go M.empty
  - Requests that failed are saved for next time. Requests that were retried
  - this time and failed are ordered last, to ensure that we don't get stuck
  - retrying the same requests and not making progress when run again.
+ -
+ - Returns any requests that failed.
  -}
-save :: S.Set Request -> Backup ()
-save retriedfailed = do
+save :: Backup [Request]
+save = do
 	commitWorkDir
 	failed <- getState failedRequests
+	retriedfailed <- getState retriedFailed
 	let toretry = S.toList failed ++ S.toList retriedfailed
 	inRepo $ storeRetry toretry
-	unless (null toretry) $
-		error $ unlines $
-			["Backup may be incomplete; " ++ 
-				show (length toretry) ++ " requests failed:"
-			] ++ map ("  " ++) (summarizeRequests toretry) ++ 
-			[ "Run again later."
-			]
+	endState
+	return toretry
+
+showFailures :: [Request] -> IO ()
+showFailures [] = noop
+showFailures l = error $ unlines $
+	["Backup may be incomplete; " ++ 
+		show (length l) ++ " requests failed:"
+	] ++ map ("  " ++) (summarizeRequests l) ++ 
+	[ "Run again later."
+	]
 
 newState :: Git.Repo -> IO BackupState
 newState r = BackupState
 	<$> pure S.empty
 	<*> pure S.empty
+	<*> pure S.empty
 	<*> pure r
-	<*> getauth
+	<*> getAuth
+	<*> pure []
+	<*> pure Nothing
+
+endState :: Backup ()
+endState = liftIO . maybe noop catFileStop =<< getState catFileHandle
+
+getAuth :: IO (Maybe Github.GithubAuth)
+getAuth = do
+	user <- getEnv "GITHUB_USER"
+	password <- getEnv "GITHUB_PASSWORD"
+	return $ case (user, password) of
+		(Just u, Just p) -> Just $ 
+			Github.GithubBasicAuth (tobs u) (tobs p)
+		_ -> Nothing
   where
-	getauth = do
-		user <- getEnv "GITHUB_USER"
-		password <- getEnv "GITHUB_PASSWORD"
-		return $ case (user, password) of
-			(Just u, Just p) -> Just $ 
-				Github.GithubBasicAuth (tobs u) (tobs p)
-			_ -> Nothing
 	tobs = encodeUtf8 . T.pack
 
+genBackupState :: Git.Repo -> IO BackupState
+genBackupState repo = newState =<< Git.Config.read repo
+
 backupRepo :: (Maybe Git.Repo) -> IO ()
 backupRepo Nothing = error "not in a git repository, and nothing specified to back up"
-backupRepo (Just repo) = evalStateT (runBackup go) =<< newState =<< Git.Config.read repo
+backupRepo (Just repo) = 
+	genBackupState repo >>= evalStateT (runBackup go) >>= showFailures
   where
 	go = do
-		retriedfailed <- retry
-		remotes <- gitHubPairs <$> getState gitRepo
-		when (null remotes) $
-			error "no github remotes found"
-		forM_ remotes $ \(r, remote) -> do
-			_ <- fetchRepo r
-			gatherMetaData remote
-		save retriedfailed
+		retry
+		mainBackup
+		runDeferred
+		save
+
+mainBackup :: Backup ()
+mainBackup = do
+	remotes <- gitHubPairs <$> getState gitRepo
+	when (null remotes) $
+		error "no github remotes found"
+	forM_ remotes $ \(r, remote) -> do
+		void $ fetchRepo r
+		gatherMetaData remote
+
+runDeferred :: Backup ()
+runDeferred = go =<< getState deferredBackups
+  where
+	go [] = noop
+	go l = do
+		changeState $ \s -> s { deferredBackups = [] }
+		void $ sequence l
+		-- Running the deferred actions could cause
+		-- more actions to be deferred; run them too.
+		runDeferred
 
 backupName :: String -> IO ()
 backupName name = do
+	auth <- getAuth
 	l <- sequence
-	 	[ Github.userRepos name Github.All
-		, Github.reposWatchedBy name
-		, Github.organizationRepos name
+	 	[ Github.userRepos' auth name Github.All
+		, Github.reposWatchedBy' auth name
+		, Github.reposStarredBy auth name
+		, Github.organizationRepos' auth name
 		]
-	let repos = concat $ rights l
-	when (null repos) $
+	let nameurls = nub $ map (\repo -> (Github.repoName repo, Github.repoGitUrl repo)) $ concat $ rights l
+	when (null nameurls) $
 		if (null $ rights l)
 			then error $ unlines $ "Failed to query github for repos:" : map show (lefts l)
 			else error $ "No GitHub repositories found for " ++ name
-	status <- forM repos $ \repo -> do
-		let dir = Github.repoName repo
+	-- Clone any missing repos, and get a BackupState for each repo
+	-- that is to be backed up.
+	states <- forM nameurls $ \(dir, url) -> do
 		unlessM (doesDirectoryExist dir) $ do
 			putStrLn $ "New repository: " ++ dir
 			ok <- boolSystem "git"
 				[ Param "clone"
-				, Param (Github.repoGitUrl repo)
+				, Param url
 				, Param dir
 				]
 			unless ok $ error "clone failed"
-		try (backupRepo . Just =<< Git.Construct.fromPath dir)
-			:: IO (Either SomeException ())
-	unless (null $ lefts status) $
-		error "Failed to successfully back everything up. Run again later."
+		genBackupState =<< Git.Construct.fromPath dir
+	-- First pass only retries things that failed before, so the
+	-- retried actions will run in each repo before too much API is
+	-- used up.
+	states' <- forM states (execStateT . runBackup $ retry)
+	states'' <- forM states' (execStateT . runBackup $ mainBackup)
+	forM states'' (evalStateT . runBackup $ runDeferred >> save)
+		>>= showFailures . concat
 
 usage :: String
 usage = "usage: github-backup [username|organization]"
